use clap::command;
use icepick_module::help::*;
use keyfork_derive_util::{request::DerivationAlgorithm, DerivationIndex, DerivationPath};
use serde::{Deserialize, Serialize};
use std::{
    collections::HashMap,
    io::{IsTerminal, Write},
    process::{Command, Stdio},
};

mod workflow;

pub fn get_command(bin_name: &str) -> (&str, Vec<&str>) {
    if std::env::vars().any(|(k, _)| &k == "ICEPICK_USE_CARGO") {
        ("cargo", vec!["run", "-q", "--bin", bin_name, "--"])
    } else {
        (bin_name, vec![])
    }
}

pub fn derive_keys(
    algo: &DerivationAlgorithm,
    path_prefix: &DerivationPath,
    accounts: &[DerivationIndex],
) -> Vec<Vec<u8>> {
    if accounts.is_empty() {
        return vec![];
    }
    let mut derived_keys = vec![];
    let mut client = keyforkd_client::Client::discover_socket().expect("keyforkd started");
    for account in accounts {
        let request = keyfork_derive_util::request::DerivationRequest::new(
            algo.clone(),
            &path_prefix.clone().chain_push(account.clone()),
        );
        let request = keyforkd_models::Request::Derivation(request);
        let response = client.request(&request).expect("valid derivation");
        match response {
            keyforkd_models::Response::Derivation(
                keyfork_derive_util::request::DerivationResponse { data, .. },
            ) => {
                derived_keys.push(data.to_vec());
            }
            _ => panic!("Unexpected response"),
        }
    }
    derived_keys
}

#[derive(Serialize, Deserialize, Debug)]
struct ModuleConfig {
    /// The name of the module.
    name: String,

    /// The name of the command used to invoke the module. If not given, the default would be
    /// `format!("icepick-{name}")`, using the name of the module.
    command_name: Option<String>,

    /// The bip32 derivation algorithm. This is currently used for deriving keys from Keyfork, but
    /// may be passed to modules within the workflow to provide additional context, such as the
    /// algorithm for a generic signer.
    #[serde(default)]
    algorithm: Option<DerivationAlgorithm>,

    /// The bip44 derivation prefix. This is currently used for deriving keys from Keyfork directly
    /// within Icepick, but may be passed to modules within the workflow to provide additional
    /// context, such as a module for deriving keys.
    #[serde(with = "serde_derivation", default)]
    derivation_prefix: Option<DerivationPath>,

    /// All workflows for a module.
    #[serde(rename = "workflow", default)]
    workflows: Vec<workflow::Workflow>,
}

mod serde_derivation {
    use keyfork_derive_util::DerivationPath;
    use serde::{Deserialize, Deserializer, Serializer};
    use std::str::FromStr;

    pub fn serialize<S>(p: &Option<DerivationPath>, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        if let Some(p) = p {
            let path = p.to_string();
            serializer.serialize_str(&path)
        } else {
            serializer.serialize_none()
        }
    }

    pub fn deserialize<'de, D>(deserializer: D) -> Result<Option<DerivationPath>, D::Error>
    where
        D: Deserializer<'de>,
    {
        use serde::de::Error;

        let opt_string = Option::<String>::deserialize(deserializer)?;
        opt_string
            .map(|string| DerivationPath::from_str(&string).map_err(Error::custom))
            .transpose()
    }
}

#[derive(Serialize, Deserialize, Debug, Default)]
struct Config {
    #[serde(rename = "module")]
    modules: Vec<ModuleConfig>,
}

// command name, invocable binary, operations
type Commands<'a> = &'a [(String, String, Vec<Operation>)];

pub fn do_cli_thing() {
    /* parse config file to get module names */
    let config_file = std::env::vars().find_map(|(k, v)| {
        if k == "ICEPICK_CONFIG_FILE" {
            return Some(v);
        }
        None
    });
    let config_path = config_file.unwrap_or_else(|| "icepick.toml".to_string());
    let config_content = std::fs::read_to_string(config_path).expect("can't read config file");
    let mut config: Config = toml::from_str(&config_content).expect("config file had invalid toml");
    config.modules.push(ModuleConfig {
        name: "internal".to_string(),
        command_name: Default::default(),
        algorithm: Default::default(),
        derivation_prefix: Default::default(),
        workflows: Default::default(),
    });

    let mut commands = vec![];
    let mut icepick_command = command!();
    for module in &config.modules {
        let module_name = &module.name;
        let bin = module
            .command_name
            .clone()
            .unwrap_or_else(|| format!("icepick-{module_name}"));
        let (command, args) = get_command(&bin);
        let mut child = Command::new(command)
            .args(args)
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .spawn()
            .unwrap();
        let mut input = child.stdin.take().unwrap();
        input
            .write_all("{\"operation\": \"help\"}\n".as_bytes())
            .unwrap();
        input
            .write_all("{\"operation\": \"exit\"}\n".as_bytes())
            .unwrap();
        let output = child.wait_with_output().unwrap().stdout;
        let operations: Vec<Operation> = serde_json::from_slice::<Vec<Operation>>(&output)
            .expect("successful deserialization of operation");
        commands.push((module_name.clone(), bin, operations));
    }

    // Add workflow subcommands
    let mut workflows = vec![];
    for module in &config.modules {
        workflows.push((module.name.clone(), module.workflows.clone()));
    }
    let workflows = workflows.leak();
    let mut workflow_command = clap::Command::new("workflow")
        .about("Run a pre-defined Icepick workflow")
        .arg(clap::arg!(--"simulate-workflow").global(true));
    for module in workflows.iter() {
        let mut module_subcommand = clap::Command::new(module.0.as_str());
        for workflow in &module.1 {
            module_subcommand = module_subcommand.subcommand(workflow.generate_command());
        }
        workflow_command = workflow_command.subcommand(module_subcommand);
    }

    icepick_command = icepick_command.subcommand(workflow_command);

    // Add per-module subcommands
    let commands = commands.leak();
    for command in commands.iter() {
        let mut subcommand = clap::Command::new(command.0.as_str());
        for op in &command.2 {
            let mut op_command =
                clap::Command::new(op.name.replace('_', "-")).about(&op.description);
            for arg in &op.arguments {
                let mut op_arg =
                    clap::Arg::new(arg.name.replace('_', "-")).help(arg.description.as_str());
                op_arg = match arg.r#type {
                    ArgumentType::Required => op_arg.required(true),
                    ArgumentType::Optional => {
                        op_arg.required(false).long(arg.name.replace('_', "-"))
                    }
                };
                op_command = op_command.arg(op_arg);
            }
            subcommand = subcommand.subcommand(op_command);
        }
        icepick_command = icepick_command.subcommand(subcommand);
    }

    let matches = icepick_command.get_matches();

    // If we have a Workflow command, run the workflow and exit.
    if let Some(("workflow", matches)) = matches.subcommand() {
        let (module_name, matches) = matches
            .subcommand()
            .expect("icepick workflow: missing module");
        let (workflow_name, matches) = matches
            .subcommand()
            .expect("icepick workflow: missing workflow");
        let workflow = workflows
            .iter()
            .find(|(module, _)| module == module_name)
            .and_then(|(_, workflows)| workflows.iter().find(|x| x.name == workflow_name))
            .expect("workflow from CLI should match config");
        workflow.handle(matches, commands, &config.modules);
        return;
    }

    // Load the input for modules if we don't have a terminal input.
    let stdin = std::io::stdin();
    let mut cli_input: Option<serde_json::Value> = None;
    if !stdin.is_terminal() {
        cli_input = serde_json::from_reader(stdin).ok();
    }
    // TODO: should we rename "blob"?
    let blob = cli_input.as_ref().and_then(|json| json.get("blob"));

    let derivation_accounts = cli_input
        .as_ref()
        .and_then(|json| json.get("derivation_accounts"));

    // We don't have a Workflow command, so let's find the matching Module command
    if let Some((module, matches)) = matches.subcommand() {
        if let Some((subcommand, matches)) = matches.subcommand() {
            if let Some(operation) = commands
                .iter()
                .find(|(name, ..)| *name == module)
                .and_then(|(.., operations)| operations.iter().find(|o| o.name == subcommand))
            {
                // all values to be passed to the command
                let mut args =
                    HashMap::<String, serde_json::Value>::with_capacity(operation.arguments.len());

                // add the values from piped input, if any
                if let Some(serde_json::Value::Object(blob)) = blob {
                    args.extend(blob.iter().map(|(k, v)| (k.clone(), v.clone())));
                }

                // add the values from CLI arguments
                for arg in &operation.arguments {
                    if let Some(value) = matches.get_one::<String>(&arg.name.replace('_', "-")) {
                        args.insert(arg.name.clone(), serde_json::Value::String(value.clone()));
                    }
                }

                let (algo, path) = config
                    .modules
                    .iter()
                    .find_map(|fmodule| {
                        if fmodule.name == module {
                            return Some((
                                fmodule.algorithm.clone(),
                                fmodule.derivation_prefix.clone(),
                            ));
                        }
                        None
                    })
                    .unwrap();

                let mut derived_keys: Vec<Vec<u8>> = vec![];
                if let Some(accounts) = derivation_accounts {
                    let accounts: Vec<keyfork_derive_util::DerivationIndex> =
                        serde_json::from_value(accounts.clone())
                            .expect("valid derivation_accounts");
                    derived_keys.extend(derive_keys(
                        &algo.expect("a module requested keys but didn't provide algorithm"),
                        &path.expect("a module requested keys but didn't provide prefix"),
                        &accounts,
                    ));
                }

                let json = serde_json::json!({
                    "operation": subcommand,
                    "values": args,
                    "derived_keys": derived_keys,
                });
                let bin = commands
                    .iter()
                    .find_map(|(fmodule, fcommand, _)| {
                        if *fmodule == module {
                            Some(fcommand)
                        } else {
                            None
                        }
                    })
                    .expect("previously found module should exist in new search");
                let (command, args) = get_command(bin);
                let mut child = Command::new(command)
                    .args(args)
                    .stdin(Stdio::piped())
                    .stdout(Stdio::piped())
                    .spawn()
                    .unwrap();
                let mut input = child.stdin.take().unwrap();
                serde_json::to_writer(&mut input, &json).unwrap();
                input.write_all(b"\n{\"operation\": \"exit\"}\n").unwrap();
                let output = child.wait_with_output().unwrap().stdout;
                let json: serde_json::Value = serde_json::from_slice(&output).expect("valid json");
                let json_as_str = serde_json::to_string(&json).unwrap();
                println!("{json_as_str}");
            }
        }
    }
}
